#include "driver_conf.h"

char **regexp(char *pattern, char *str)
{

	char **rm = NULL;
	int result = 0;
	int i = 1;

	regex_t re;
	regmatch_t *rem;

	char *buffer;
	int length;

	result = regcomp(&re, pattern, REG_EXTENDED);

	if (result != 0) {
		PERROR("error in regcomp");
		return NULL;
	}

	rem = malloc((re.re_nsub + 1) * sizeof(regmatch_t));
	if (rem == NULL)
		return NULL;

	result = regexec(&re, str, re.re_nsub + 1, rem, 0);

	if (result != 0) {
		if (result == REG_NOMATCH)
			PERROR("REG_NOMATCH");
		else
			PERROR("error in regexec");
		free(rem);
		regfree(&re);
		return NULL;
	}

	rm = malloc(re.re_nsub * sizeof(char *));	/* should free */
	if (rm == NULL)
		return NULL;

	for (i = 1; i < (re.re_nsub + 1); i++) {
		length = rem[i].rm_eo - rem[i].rm_so;
		buffer = malloc((length + 1) * sizeof(char));	/* one day we should free these ones */
		strncpy(buffer, str + rem[i].rm_so, length);
		buffer[length] = 0;

		rm[i - 1] = buffer;
	}

	free(rem);
	regfree(&re);

	return rm;
}



int set_fault_policy(driver_t * d, char * str)
{
	char **rm;
	command_t * cmd = &(d->fault_cmds[d->nr_fault_cmds]);

	if (strncmp(str, "always recover\n", 14) == 0) {
		cmd->type = POL_RECOVER;
		d->nr_fault_cmds++;

		writelog("\tadding policy POL_RECOVER\n");
		return 0;
	}

	if (strncmp(str, "do not recover\n", 14) == 0) {
		cmd->type = POL_DO_NOT_RECOVER;
		d->nr_fault_cmds++;

		writelog("\tadding policy POL_DO_NOT_RECOVER\n");
		return 0;
	}

	rm = regexp ("recover ([[:digit:]]+) times in ([[:digit:]]+) minutes", str);
	if (rm != NULL) {
		cmd->type = POL_RECOVER_N_M;
		cmd->int1 = atoi(rm[0]);
		cmd->int2 = atoi(rm[1]);
		d->nr_fault_cmds++;

		writelog("\tadding policy POL_RECOVER_N_M %d, %d type=%x\n",
			 cmd->int1, cmd->int2, cmd->type);
		free(rm[0]); free(rm[1]); free(rm);	/* free the matches of the last regexp */
		return 0;
	}

	rm = regexp("recover ([[:digit:]]+) times", str);
	if (rm != NULL) {
		cmd->type = POL_RECOVER_N;
		cmd->int1 = atoi(rm[0]);
		d->nr_fault_cmds++;

		writelog("\tadding policy POL_RECOVER_N %d\n",
			 cmd->int1);
		free(rm[0]); free(rm);	/* free the matches of the last regexp */
		return 0;
	}

	rm = regexp("replace module by (.+)\n", str);
	if (rm != NULL) {
		cmd->type = CMD_REPLACE;
		cmd->char1 = rm[0];
		d->nr_fault_cmds++;

		writelog("\tadding policy CMD_REPLACE\n");
		free(rm);	/* free the matches of the last regexp but not
				   the first one used to store the name of the module */
		return 0;
	}

	return -1;
}


int add_command(driver_t * d, int type, char *str)
{
	char **rm;
	command_t * cmd;
	int *nr;

	switch (type) {
	case ON_FAULT:
		if (set_fault_policy(d, str) == 0) return 0;
		cmd = d->fault_cmds;
		nr = &d->nr_fault_cmds;
		break;
	case ON_RECOVER:
		cmd = &(d->recover_cmds[d->nr_recover_cmds]);
		nr = &d->nr_recover_cmds;
		break;
	case ON_LAST:
		cmd = &(d->last_cmds[d->nr_last_cmds]);
		nr = &d->nr_last_cmds;
		break;
	default:
		return -1;

	}

	rm = regexp("execute (.+)", str);
	if (rm != NULL) {
		cmd->type = CMD_EXECUTE;
		cmd->char1 = rm[0];
		*nr = *nr + 1;

		writelog("\tadding command CMD_EXECUTE\n");
		free(rm);	/* free the matches of the last regexp */
		return 0;
	}

	rm = regexp("send email to (.+)", str);
	if (rm != NULL) {
		cmd->type = CMD_MAIL;
		cmd->char1 = rm[0];
		*nr = *nr + 1;

		writelog("\tadding command CMD_MAIL\n");
		free(rm);	/* free the matches of the last regexp */
		return 0;
	}

	writelog("invalid command\n");
	return -1;
}

int parse_commands(driver_t * d, int type, char *str)
{

	char **rm;

	rm = regexp("(.+) and (.+)", str);

	if (rm != NULL) {
		parse_commands(d, type, rm[0]);
		add_command(d, type, rm[1]);
		free(rm[0]);
		free(rm[1]);
		free(rm);
		return 0;
	}

	return add_command(d, type, str);
}

int parse_line(driver_t * d, char *line)
{
	int type = ON_FAULT;
	char **rm;

	rm = regexp("on ([[:alpha:]]+) (.+)\n", line);

	if (rm != NULL) {
		if (strncmp(rm[0], "fault", 5) == 0) {
			parse_commands(d, ON_FAULT, rm[1]);
		} else if (strncmp(rm[0], "recover", 7) == 0) {
			type = ON_RECOVER;
			parse_commands(d, ON_RECOVER, rm[1]);
		} else if (strncmp(rm[0], "last", 7) == 0) {
			type = ON_LAST;
			parse_commands(d, ON_LAST, rm[1]);
		} else {
			return -1;
		}

		free(rm[0]);
		free(rm[1]);
		free(rm);

		flush_log();
		return 0;
	}

	return -1;
}

int load_driver_conf(driver_t * d)
{
	FILE *fd;
	char buffer[1024];
	int result = 0;

	d->nr_fault_cmds = 0;

	if (d->name == NULL) {
		writelog("nooks-agent: driver name missing\n");
	}

	sprintf(buffer, NOOKS_MODULE_PATH "%s.conf", d->name);

	fd = fopen(buffer, "r");

	/* affect default policy to driver */
	d->fault_cmds[d->nr_fault_cmds].type = POL_RECOVER;
	if (fd == NULL) {
		writelog(" can't open file \"%s\".\n", buffer);
		writelog(" Affecting default policy to driver %s.\n",
			 d->name);
		d->nr_fault_cmds++;
		return 0;
	}

	while (fgets(buffer, 1023, fd)) {
		result = parse_line(d, buffer);
		if (result < 0) {
			writelog(" error fgets in %s\n", __FUNCTION__);
			flush_log();
		}
	}

	return 0;
}

int execute_driver_cmds(driver_t * d, int type)
{
	int i;
	switch(type) {
		case ON_FAULT:
			for (i = 0; i < d->nr_fault_cmds; i++) {

				if (d->fault_cmds[i].type == CMD_EXECUTE) {
					writelog(" CMD_EXECUTE ...\n");
					flush_log();

					execute_command(d->fault_cmds[i].char1);
				} else if (d->fault_cmds[i].type == CMD_MAIL) {
					writelog(" CMD_MAIL ...\n");
					flush_log();

					send_mail_fault(d->fault_cmds[i].char1);
				}

			}
			break;
		case ON_RECOVER:
			for (i = 0; i < d->nr_recover_cmds; i++) {

				if (d->recover_cmds[i].type == CMD_EXECUTE) {
					writelog(" CMD_EXECUTE ...\n");
					flush_log();

					execute_command(d->recover_cmds[i].char1);
				} else if (d->recover_cmds[i].type == CMD_MAIL) {
					writelog(" CMD_MAIL ...\n");
					flush_log();

					send_mail_recover(d->recover_cmds[i].char1);
				}

			}
		break;
		case ON_LAST:
			for (i = 0; i < d->nr_last_cmds; i++) {

				if (d->last_cmds[i].type == CMD_EXECUTE) {
					writelog(" CMD_EXECUTE ...\n");
					flush_log();

					execute_command(d->last_cmds[i].char1);
				} else if (d->last_cmds[i].type == CMD_MAIL) {
					writelog(" CMD_MAIL ...\n");
					flush_log();

					send_mail_recover(d->last_cmds[i].char1);
				}

			}
		break;
	}
	return 0;
}

int do_driver_recover(driver_t * d)
{
	int i = 0;
	command_t * cmd;
	int result;
	
	for (i = 0; i < d->nr_fault_cmds; i++) {
		if (IS_POL(d->fault_cmds[i].type)) break;
	}

	cmd = &(d->fault_cmds[i]);
	
	switch (cmd->type) {
	case POL_RECOVER:
		return 1;
		break;
	case POL_DO_NOT_RECOVER:
		return 0;
		break;
	case POL_RECOVER_N:
		result = get_times(d->name);
			writelog(" The driver has already recovered %d times\n", result);
		if (result < cmd->int1) {
			return 1;
		} else {
			return 0;
		}
		break;
	case POL_RECOVER_N_M:
		result = get_times_in_minutes(d->name, cmd->int2);
			writelog(" The driver has already recovered %d times in %d minutes\n",
					result, cmd->int2);
		if (result < cmd->int1) {
			return 1;
		} else {
			return 0;
		}
		break;
	default:
	}
	return 0;

}
